/* 
 * Shared library add-on to iptables to match 
 * packets by the incoming interface group.
 *
 * (c) 2006, 2007 Balazs Scheidler <[EMAIL PROTECTED]>,
 * Laszlo Attila Toth <[EMAIL PROTECTED]>
 */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <xtables.h>
#include <linux/netfilter/xt_ifgroup.h>

/*
modification history
--------------------
01a,05nov08,sam ported from iptables from xtables
*/

static void
ifgroup_help(void)
{
       printf(
"ifgroup match options:\n"
"  [!] --ifgroup-in  group[/mask]  incoming interface group and its mask\n"
"  [!] --ifgroup-out group[/mask]  outgoing interface group and its mask\n"
"  --tproxy-mask mask              tproxy mask\n"
"  --nf-hook hook                  netfilter hook\n"
"\n");
}

static const struct option ifgroup_opts[] = {
       {"ifgroup-in", 1, NULL, '1'},
       {"ifgroup-out", 1, NULL, '2'},
       {"tproxy-mask", 1, NULL, '3'},
       {"nf-hook", 1, NULL, '4'},       
       { }
};

#define PARAM_MATCH_IN         0x01
#define PARAM_MATCH_OUT        0x02
#define PARAM_MATCH_MASK       0x04
#define PARAM_NF_HOOK          0x08


#define IFGROUP_DEFAULT_MASK 0xffffffffU

static void parse_nf_hook(char *s, struct xt_ifgroup_info *info)
{
        u_int32_t tmp;

        if (xtables_strtoui(s, NULL, &tmp, 0, UINT8_MAX) < 0) {
                xtables_error(PARAMETER_PROBLEM,
                              "bad value in --nf-hook \"%s\"", s);
        }

        if (tmp > XT_IFGROUP_NF_MAX) {
                xtables_error(PARAMETER_PROBLEM,
                              "bad value in --nf-hook \"%s\"", s);
        }

        info->nf_hook = (u_int8_t)tmp;
}

static int
ifgroup_parse(int c, char **argv, int invert, unsigned int *flags,
              const void *entry, struct xt_entry_match **match)
{
       struct xt_ifgroup_info *info =
                        (struct xt_ifgroup_info *) (*match)->data;
       char *end;

       switch (c) {
       case '1':
               if (*flags & PARAM_MATCH_IN)
                       xtables_error(PARAMETER_PROBLEM,
                           "ifgroup match: Can't specify --ifgroup-in twice");

               xtables_check_inverse(optarg, &invert, &optind, 0, argv);

               info->in_group = strtoul(optarg, &end, 0);
               info->in_mask = IFGROUP_DEFAULT_MASK;

               if (*end == '/')
                       info->in_mask = strtoul(end+1, &end, 0);

               if (*end != '\0' || end == optarg)
                       xtables_error(PARAMETER_PROBLEM,
                                 "ifgroup match: Bad ifgroup value `%s'",
                                 optarg);

               if (invert)
                       info->flags |= XT_IFGROUP_INVERT_IN;

               *flags |= PARAM_MATCH_IN;
               info->flags |= XT_IFGROUP_MATCH_IN;
               break;

       case '2':
               if (*flags & PARAM_MATCH_OUT)
                       xtables_error(PARAMETER_PROBLEM,
                           "ifgroup match: Can't specify --ifgroup-out twice");

               xtables_check_inverse(optarg, &invert, &optind, 0, argv);

               info->out_group = strtoul(optarg, &end, 0);
               info->out_mask = IFGROUP_DEFAULT_MASK;

               if (*end == '/')
                       info->out_mask = strtoul(end+1, &end, 0);

               if (*end != '\0' || end == optarg)
                       xtables_error(PARAMETER_PROBLEM,
                           "ifgroup match: Bad ifgroup value `%s'", optarg);

               if (invert)
                       info->flags |= XT_IFGROUP_INVERT_OUT;

               *flags |= PARAM_MATCH_OUT;
               info->flags |= XT_IFGROUP_MATCH_OUT;
               break;

       case '3':                
               if (*flags & PARAM_MATCH_MASK)
                       xtables_error(PARAMETER_PROBLEM,
                           "ifgroup match: Can't specify --tproxy-mask twice");
           
               xtables_check_inverse(optarg, &invert, &optind, 0, argv);
               info->tproxy_mark_mask = strtoul(optarg, &end, 0);
               if (invert) info->flags |= XT_IFGROUP_INVERT_MASK;

               *flags |= PARAM_MATCH_MASK;
               info->flags |= XT_IFGROUP_MATCH_MASK;               
               break;

       case '4':
	       if (*flags & PARAM_NF_HOOK) {
		       xtables_error(PARAMETER_PROBLEM,
			   "ifgroup match: Can't specify --nf-hook twice");
	       }

	       parse_nf_hook(optarg, info);
	       *flags |= PARAM_NF_HOOK;
	       info->flags |= XT_IFGROUP_MATCH_NF_HOOK;
	       break;

       default: 
               return 0;
       }

       return 1;
}

static void
ifgroup_final_check(unsigned int flags)
{
       if (!flags)
               xtables_error(PARAMETER_PROBLEM,
                   "You must specify either "
                   "`--ifgroup-in' or `--ifgroup-out'");
}

static void
ifgroup_print_value_in(struct xt_ifgroup_info *info)
{
       printf("0x%x", info->in_group);
       if (info->in_mask != IFGROUP_DEFAULT_MASK)
               printf("/0x%x", info->in_mask);
       printf(" ");
}

static void
ifgroup_print_value_out(struct xt_ifgroup_info *info)
{
       printf("0x%x", info->out_group);
       if (info->out_mask != IFGROUP_DEFAULT_MASK)
               printf("/0x%x", info->out_mask);
       printf(" ");
}

static void
ifgroup_print_value_mask(struct xt_ifgroup_info *info)
{
       printf("0x%x", info->tproxy_mark_mask);
       printf(" ");
}

ifgroup_print_value_nf_hook(struct xt_ifgroup_info *info)
{
        printf("0x%x", info->nf_hook);
        printf(" ");
}

static void
ifgroup_print(const void *ip,
             const struct xt_entry_match *match,
             int numeric)
{
       struct xt_ifgroup_info *info =
               (struct xt_ifgroup_info *) match->data;

       printf("ifgroup ");

       if (info->flags & XT_IFGROUP_MATCH_IN) {
               printf("in %s",
                      info->flags & XT_IFGROUP_INVERT_IN ? "! " : "");
               ifgroup_print_value_in(info);
       }
       if (info->flags & XT_IFGROUP_MATCH_OUT) {
               printf("out %s",
                      info->flags & XT_IFGROUP_INVERT_OUT ? "! " : "");
               ifgroup_print_value_out(info);
       }
       if (info->flags & XT_IFGROUP_MATCH_MASK) {
               printf("tproxy mask %s",
                      info->flags & XT_IFGROUP_INVERT_MASK ? "! " : "");
               ifgroup_print_value_mask(info);
       }       

       if (info->flags & XT_IFGROUP_MATCH_NF_HOOK) {
	       printf("nf-hook 0x%x ", info->nf_hook);
	       ifgroup_print_value_nf_hook(info);
       }
}

static void
ifgroup_save(const void *ip, const struct xt_entry_match *match)
{
       struct xt_ifgroup_info *info =
               (struct xt_ifgroup_info *) match->data;

       if (info->flags & XT_IFGROUP_MATCH_IN) {
               printf("%s--ifgroup-in ",
                      info->flags & XT_IFGROUP_INVERT_IN ? "! " : "");
               ifgroup_print_value_in(info);
       }
       if (info->flags & XT_IFGROUP_MATCH_OUT) {
               printf("%s--ifgroup-out ",
                      info->flags & XT_IFGROUP_INVERT_OUT ? "! " : "");
               ifgroup_print_value_out(info);
       }
       if (info->flags & XT_IFGROUP_MATCH_MASK) {
               printf("%s--tproxy-mask ",
                      info->flags & XT_IFGROUP_INVERT_MASK ? "! " : "");
               ifgroup_print_value_mask(info);
       }       
       if (info->flags & XT_IFGROUP_MATCH_NF_HOOK) {
	       printf("--nf-hook ");
           ifgroup_print_value_nf_hook(info);
       }       
}

static struct xtables_match ifgroup_match = {
        .family		= NFPROTO_UNSPEC,
        .name           = "ifgroup",
        .version        = XTABLES_VERSION,
        .size           = XT_ALIGN(sizeof(struct xt_ifgroup_info)),
        .userspacesize  = XT_ALIGN(sizeof(struct xt_ifgroup_info)),
        .help           = ifgroup_help,
        .parse          = ifgroup_parse,
        .final_check    = ifgroup_final_check,
        .print          = ifgroup_print,
        .save           = ifgroup_save,
        .extra_opts     = ifgroup_opts
};

void _init(void)
{
       xtables_register_match(&ifgroup_match);
}
